#include <strings.h>
#include <math.h>
#include "particulas.h"

#define MAX(A,B)       ((A)>(B))?(A):(B)
#define MIN(A,B)       ((A)<(B))?(A):(B)
#define RANDBASE       (((double)random())/(RAND_MAX))
#define RAND(INF,SUP)  (((RANDBASE)*((SUP)-(INF)))+(INF))

/* Funcion auxiliar de movimiento en busqueda de una posicion,
   considerando la velocidad actual de la particula */
void buscar(particula_t* o, vector_t* pos, vector_t* fuerza)
{
  copia(fuerza,pos);
  resta(fuerza,&(o->posicion));
  normalizar(fuerza);
  producto_escalar(fuerza,MAX_VELOCIDAD);
  resta(fuerza,&(o->velocidad));
}

/* Determina si dos particulas se encuentran en el radio de 
   influencia mutua (son vecinas) */
int vecino(particula_t* o1, particula_t* o2, double radio)
{
  vector_t v;

  copia(&v,&(o1->posicion));
  resta(&v,&(o2->posicion));
  return (longitud(&v)<radio);
}

/* Determina si una particula tiene vecinos en su radio de accion, si no
   es asi no se ve modificada por las fuerzas de grupo */
int tiene_vecinos(particula_t* o, particula_t* vo, size_t n_objs, double radio)
{
  size_t i;

  for(i=0;i<n_objs;i++)
    if(vecino(o,&(vo[i]),radio))
      return 1;

  return 0;
}

/* Fuerza de separacion: Es la agregacion de fuerzas que le separan de cada 
   una de las particulas con las que forma grupo (vecinas), esta fuerza es
   proporcional a la distancia a cada particula */
void separacion(particula_t* o, particula_t* vo, size_t n_objs, double radio, vector_t* fuerza)
{
  size_t i;
  
  double dist;
  vector_t aux;

  cero(fuerza);

  for(i=0;i<n_objs;i++)
    if(o!=&(vo[i]) && vecino(o,&(vo[i]),radio))
      {
	copia(&aux,&(o->posicion));
	resta(&aux,&(vo[i].posicion));
	dist=longitud(&aux);
	normalizar(&aux);
	if(dist<0.001)
	  dist=0.001;
	producto_escalar(&aux,1/dist);
	suma(fuerza,&aux);
      }
}

/* Funcion de control de la entropia: Para las particulas de un grupo (vecinas) se
   determina el grado de entropia, como la distancia de cada una de las particulas
   del grupo al resto por la velocidad de la particula.
   - Si el numero de particulas en el grupo esta por debajo de la masa critica
     la fuerza reduce la entropia, ejerciendo una fuerza de frenado, opuesta
     a la velocidad actual y proporcional al inverso de la entropia.
   - Si se excede el umbral de masa critica la fuerza acelera a la particula
     prpporcionalmente a la entropia. */
void entropia(particula_t* o, particula_t* vo, size_t n_objs, double radio, vector_t* fuerza)
{
  double entropia, parcial;
  vector_t distancia;
  size_t i,j;
  size_t cnt=0;

  for(i=0;i<n_objs;i++)
    if(o!=&(vo[i]) && vecino(o,&(vo[i]),radio))
      {
	cnt++;
	parcial=0.0;
	for(j=0;j<n_objs;j++)
	  if(&(vo[i])!=&(vo[j]) && vecino(o,&(vo[j]),radio))
	    {
	      copia(&distancia,&(vo[i].posicion));
	      resta(&distancia,&(vo[j].posicion));
	      parcial+=longitud(&distancia);
	    }
	entropia+=CONST_ENTROPIA*(parcial*longitud(&(vo[i].velocidad)));
      }

  if(entropia<0.001)
    entropia=0.001;
  
  copia(fuerza,&(o->velocidad));
  if(cnt<MASA_CRITICA)
    producto_escalar(fuerza,-(1.0/entropia));
  else
    producto_escalar(fuerza,5.0*pow(entropia,2));
}

/* Fuerza de cohesion: Tiende hacia el centro de masas de un grupo de particulas */
void cohesion(particula_t* o, particula_t* vo, size_t n_objs, double radio, vector_t* fuerza)
{
  size_t i,cnt=0;
  
  vector_t centro;

  cero(fuerza);
  cero(&centro);

  for(i=0;i<n_objs;i++)
    if(o!=&(vo[i]) && vecino(o,&(vo[i]),radio))
      {
	suma(&centro,&(vo[i].posicion));
	cnt++;
      }

  if(cnt>0)
    {
      producto_escalar(&centro,1.0/cnt);
      buscar(o,&centro,fuerza);
    }  
}

/* Fuerza de dispersion: Incluye una pequena dispersion aleatoria */
void dispersion(vector_t* fuerza, double rango) 
{
  fuerza->x+=RAND(-rango,rango);
  fuerza->y+=RAND(-rango,rango);
}

/* Fuerza de rotacion: Incluye una fuerza de rotacion sobre el pto (0,0) */
void rotacion(particula_t* o, vector_t* fuerza) 
{
  double dist;
  fuerza->x= o->posicion.y;
  fuerza->y=-o->posicion.x;

  dist=longitud(fuerza);
  producto_escalar(fuerza,1/dist);
}

/* Funcion auxiliar para imprimir un vector */
void imprimir_v(vector_t* v)
{
  fprintf(stdout,"Vector: %lf, %lf\n",v->x,v->y);
}

/* Fuerza de repulsion: Fuerza opuesta hacia un conjunto de elementos llamados repulsores */
void repulsion(particula_t* o, particula_t* vr, size_t n_reps, vector_t* fuerza)
{
  size_t i;
  vector_t dir;
  double dist;

  cero(fuerza);

  for(i=0;i<n_reps;i++)
    {
      copia(&dir,&(o->posicion));
      resta(&dir,&(vr[i].posicion));
      dist=longitud(&dir);
      if(dist<REPULSION_DIST)
	{
	  if(dist<0.001)
	    dist=0.001;
	  producto_escalar(&dir,1/dist);
	  suma(fuerza,&dir);
	}
    }
}

/* Funcion de combinacion de todas las fuerzas */
void combinar(particula_t* o, particula_t* vo, size_t n_objs, particula_t* vr, size_t n_reps, vector_t* fuerza)
{
  vector_t separador, cohesionador, dispersor, repulsor, entropico, rotador;

  cero(&separador);
  cero(&cohesionador);
  cero(&dispersor);
  cero(&repulsor);
  cero(&entropico);
  cero(&rotador);

  if(tiene_vecinos(o,vo,n_objs,RADIO))
    {
      separacion(o,vo,n_objs,RADIO,&separador);
      producto_escalar(&separador,FACTOR_SEP);

      cohesion(o,vo,n_objs,RADIO,&cohesionador);
      producto_escalar(&cohesionador,FACTOR_COH);

      entropia(o,vo,n_objs,RADIO,&entropico);
      producto_escalar(&entropico,FACTOR_ENT);
    }

  dispersion(&dispersor,DISPERSION);
  producto_escalar(&dispersor,FACTOR_DIS);

  repulsion(o,vr,n_reps,&repulsor);
  producto_escalar(&repulsor,FACTOR_REP);

  rotacion(o,&rotador);
  producto_escalar(&rotador,FACTOR_ROT);

#ifdef DEBUG
  fprintf(stdout,"Separador:   "); imprimir_v(&separador);         
  fprintf(stdout,"Cohesionado: "); imprimir_v(&cohesionador);
  fprintf(stdout,"Dispersor:   "); imprimir_v(&dispersor);
  fprintf(stdout,"Repulsor:    "); imprimir_v(&repulsor);
  fprintf(stdout,"Entropico:   "); imprimir_v(&entropico);
  fprintf(stdout,"Rotador:     "); imprimir_v(&rotador);
#endif       

  cero(fuerza);
  suma(fuerza,&separador);         
  suma(fuerza,&cohesionador);
  suma(fuerza,&dispersor);
  suma(fuerza,&repulsor);
  suma(fuerza,&entropico);
  suma(fuerza,&rotador);

  truncar(fuerza,MAX_FUERZA);

#ifdef DEBUG
  fprintf(stdout,"FINAL:     "); imprimir_v(fuerza);
  fprintf(stdout,"VELOCIDAD: "); imprimir_v(&(o->velocidad));
#endif
}

/* Aplica la aceleracion debida a las fuerzas que actuan sobre la particula */
void acelerar(particula_t* o, vector_t* fuerza)
{
  vector_t aceleracion;
  copia(&aceleracion,fuerza);
  producto_escalar(&aceleracion,TIEMPO);
  suma(&(o->velocidad),&aceleracion);
}

/* Ejecuta un paso de la simulacion, moviendo la particula segun la velocidad
   que tiene. Esta funcion vigila que la particula no se salga de los limites 
   de la simulacion por medio de un fecto rebote */
void mover(particula_t* o)
{
  suma(&(o->posicion),&(o->velocidad));

  o->posicion.x=MAX(o->posicion.x,MIN_X);
  o->posicion.x=MIN(o->posicion.x,MAX_X);

  o->posicion.y=MAX(o->posicion.y,MIN_Y);
  o->posicion.y=MIN(o->posicion.y,MAX_Y);

  if(o->posicion.x==MIN_X)   o->velocidad.x= REBOTE;
  if(o->posicion.x==MAX_X)   o->velocidad.x=-REBOTE;
  if(o->posicion.y==MIN_Y)   o->velocidad.y= REBOTE;
  if(o->posicion.y==MAX_Y)   o->velocidad.y=-REBOTE;
}

/* Actualizacion de las posiciones de los repulsores */
void actualizar_repulsores(particula_t* vr, size_t n_reps, size_t min_rep, size_t max_rep)
{
  size_t i;
  vector_t fuerza;

  /* Los repulsores solo se ven afectados por las fuerzas de rotacion */
  for(i=min_rep;i<max_rep;i++)
    {
      rotacion(&(vr[i]),&fuerza);
      truncar(&fuerza,MAX_FUERZA/2.0);
      acelerar(&(vr[i]),&fuerza);
      mover(&(vr[i]));
    }
}

/* Actualizacion de las posiciones de las particulas */
void actualizar_particulas(particula_t* vo, size_t n_objs, size_t min_obj, size_t max_obj, particula_t* vr, size_t n_reps)
{
  size_t i;
  vector_t fuerza;

  for(i=min_obj;i<max_obj;i++)
    {
      combinar(&(vo[i]),vo,n_objs,vr,n_reps,&fuerza);
      acelerar(&(vo[i]),&fuerza);
    }

  /* Requiere hacerlo en dos pasadas pra que no afecta a las fuerzas 
     inciden sobre otras particulas */
  for(i=min_obj;i<max_obj;i++)
    mover(&(vo[i]));
}

/* Creacion de una nueva particula, inicialmente detenida */
void nueva_particula(particula_t* o)
{
  o->posicion.x=RAND(MIN_X,MAX_X);
  o->posicion.y=RAND(MIN_Y,MAX_Y);
  o->velocidad.x=0.0;
  o->velocidad.y=0.0;
}

/* Inicializa un conjunto de particulas/repulsores */
particula_t* inicializar(size_t n_objs)
{
  particula_t* vo;
  size_t i;

  vo=(particula_t*)malloc(sizeof(particula_t)*n_objs);
  bzero((void*)vo,sizeof(particula_t)*n_objs);

  for(i=0;i<n_objs;i++)
    nueva_particula(&(vo[i]));

  return vo;
}

/* Imprime una particula */
void imprimir_particula(particula_t* o)
{
  fprintf(stdout,
	  "%lf;%lf;%lf;%lf\n",
	  o->posicion.x,
	  o->posicion.y,
	  o->velocidad.x,
	  o->velocidad.y);
}

/* Imprime un conjunto de particulas */
void imprimir(particula_t* vo, size_t n_objs)
{
  size_t i;

  for(i=0;i<n_objs;i++)
    imprimir_particula(&(vo[i]));
}
